<?php

/*
 * This script is run on package update, from a boot script on the
 * first reboot after an update, *before* the HTTP service is started.
 * Note that stdout and stderr go to the console, not a browser. 
 *
 * Put here any complex operations that are required after a package
 * update. As much as possible the operations should only be performed
 * if necessary, so that minimal system modification is done.
 *
 * The operations in here should also be idempotent, running them more
 * than once should have no ill effects, as this may happen under some
 * circusmtances after an update (think power loss at just the wrong
 * moment).
 */

include_once dirname(__FILE__).'/SecurityJobs.php';

/* Make sure the random, local-only, "updater" passwords to access the
 * password protected admin interface and RPC service are present.
 * These are required as of firmware 2.2.5 and are added when the user
 * sets up a password, but are missing when password protection was
 * set up with a 2.2.4 or earlier firmware and then the firmware
 * updated.
 */

function createUpdaterPasswords()
{
    $job = new SecurityJobs;
    $user = $job->getUsers();
    if ( count($user) == 0 )
        return; // system is not password protected
    foreach ( $user as $u ) {
        if ( $u['user'] == 'uploader' )
            return; // uploader user already present
    }
    echo "creating local updater passwords... ";
    // there is some password protection but the random uploader
    // password is not present, saving the user passwords will add
    // them
    $job->saveUsers( $user );
    echo "done\n";
}

createUpdaterPasswords();

function updateCockpitICConfig()
{
    $spxNS = 'http://www.spinetix.com/namespace/1.0/spxconf';
    $xlinkNS = 'http://www.w3.org/1999/xlink';
    $config_name = '/etc/raperca/uploaderconf.xml';
    $uploadercal = "/etc/raperca/uploader.ics";
    
    $config = new DOMDocument;
    $config->loadXML( file_get_contents($config_name) );
    $schedule = $config->getElementsByTagName( "schedule" );
    $main = $config->documentElement;
    
    if ( $schedule->length > 0 ) {        
        $link = $schedule->item(0)->getAttributeNS($xlinkNS,"href");
        //echo $link . "\n";
        if ( $link == "uploader.ics" ){
            $link = $uploadercal;
        }
        if ( $link == $uploadercal && file_exists( $link )) {
            $ics = file_get_contents( $link );
            $matches = array();
            $fixedics = str_replace(array("\r\n", "\r"), "\n", str_replace(array("\r\n ", "\n ", "\r "), "", $ics));
            
            if ( strpos( $fixedics, "PRODID:icalendar-ruby" ) ) {
                // Find each event 
                
                $st = strpos( $fixedics, "BEGIN:VEVENT");
                while( $st !== false ){                    
                    $ed = strpos( substr($fixedics, $st), "END:VEVENT");
                    if ( $ed === false ){
                        return;
                    }
                    $event = substr($fixedics, $st, $ed+strlen("END:VEVENT") );
                    //echo "------------\n" . $event . "\n------------\n";
                    if ( strpos("SUMMARY:rpc", $event) !== null && preg_match("/URL\:(https:\/\/concentrator\.spinetix\.com.*)/", $event, $matches )) {
                        $cockpitUri = trim($matches[1]);
                        if ( preg_match("/DESCRIPTION\:(.*)/", $fixedics, $matches ) ){
                            $options = str_replace(array("\\,"), ",", $matches[1] );
                            $cockpitOptions = trim( str_replace(array("\\r\\n", "\\r", "\\n"), "\n", $options) );
                        } else {
                            $cockpitOptions = "";
                        }
                        if ( preg_match("/DTSTART\:\d*T(\d{2})(\d{2})(\d{2})/", $fixedics, $matches ) ){                        
                            $time = intval( $matches[1], 10 )* 3600 +  intval( $matches[2], 10 ) *60 + intval( $matches[3], 10 );
                            $cockpitTime = $time % 300;
                        } else {
                            $cockpitTime = rand(0, 300);
                        }
                        // kill this event from the ICS file                        
                        $fixedics = substr($fixedics, 0, $st) . substr($fixedics, $st + $ed +strlen("END:VEVENT")+1 );
                        file_put_contents($link, $fixedics );
                        // Update the config
                        $proj = $main->appendChild( $config->createElementNS($spxNS,'action') );
                        $proj->setAttribute("summary", 'rpc');
                        $proj->setAttribute("guid", uniqid() );
                        $proj->setAttribute("time", $cockpitTime );
                        $proj->setAttribute("end", 24*3600 );
                        $proj->setAttribute("cockpit", "1" );
                        $proj->setAttributeNS($xlinkNS, "xlink:href", $cockpitUri);                        
                        $proj->appendChild($config->createElementNS($spxNS, 'desc'))->appendChild( $config->createTextNode( $cockpitOptions ) );
                        file_put_contents($config_name, $config->saveXML() );
                        
                        $st = strpos( $fixedics, "BEGIN:VEVENT");
                    } else {
                        $st = strpos( substr($fixedics, $st + $ed+strlen("END:VEVENT") ), "BEGIN:VEVENT") + $st + $ed+strlen("END:VEVENT");
                    }
                }                
            }
        }
    }
}

updateCockpitICConfig();